# -*- coding: utf-8 -*-

__author__ = "Luis Barroso-Luque"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


# Threading debug notes:
# 1. YOU CANNOT COMMUNICATE NUMPY ARRAYS ACROSS THREADS ON WINDOWS BUILDS!
# 2. SOME PYFAI INTEGRATION METHODS DON'T WORK ACROSS THEADS ON WINDOWS; ONLY NUMPY/CYTHON WORK!


import sys
import types
import traceback
import functools
import Queue
import multiprocessing as mp
from PySide import QtCore
from pipeline import msg
# Error is raised if this import is removed probably due to some circular with this module and something???
from client import spot, globus, sftp
from modpkgs import nonesigmod
from modpkgs import guiinvoker

class Emitter(QtCore.QObject):
    """
    Class that holds signals that can be emitted by a QRunnable

    Attributes
    ----------
    sigRetValue : QtCore.Signal
        Signal used to emit the return value from the function passed to Runnable
    sigFinised : QtCore.Signal
        Void signal emitted when the runnable has finished correctly (no exceptions raised)
    sigExcept : QtCore.Signal
        Signal that emits the exeption type, exception instance and traceback for an exception raised when running
        the Runnable
    """

    sigRetValue = QtCore.Signal(object)
    sigFinished = QtCore.Signal()
    sigExcept = QtCore.Signal(type, Exception, types.TracebackType)



def EmitterFactory(*sig):
    return type('Emitter',(Emitter,),{'sigTemp':QtCore.Signal(*sig)})

class RunnableMethod(QtCore.QRunnable):
    """
    Runnable that will execute a given method from a QThreadPool and emit the response to the given callback function
    To use simply instantiate and pass the instance to a QThreadPool.start()


    Attributes
    ----------
    emmiter : QtCore.QObject
        QObject used to emit signals to communicate with the main GUI thread
    _method : function/method
        Function/method that will be run in a background thread
    method_args : list
        List/tuple of arguments for the given function/method
    method_kwargs : dict
        Dictionary with keyword arguments for the give function/method
    lock : object
        Mutex used to lock when several threads are spawned
    _priority : int
        Priority given to this Runnable in the pool


    Parameters
    ----------
    method : function
        Function object, method to run on seperate thread
    method_args : list
        List/tuple of arguments for the given function/method
    method_kwargs : dict
        Dictionary with keyword arguments for the give function/method
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    default_exhandle : bool
        Flag to use the default exception handle slot. If false it will not be called
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented
    """

    def __init__(self, method, method_args=(), method_kwargs={}, callback_slot=None,
                 finished_slot=None, except_slot=None, default_exhandle=True, priority=0, lock=None):
        super(RunnableMethod, self).__init__()
        self.emitter = Emitter()
        self._method = method
        self.method_args = method_args
        self.method_kwargs = method_kwargs
        self._priority = priority
        self.lock = lock

        # Connect callback and finished slots to corresponding signals
        self._callback_slot = callback_slot
        if finished_slot is not None:
            self.emitter.sigFinished.connect(finished_slot, QtCore.Qt.QueuedConnection)
        if except_slot is not None:
            self.emitter.sigExcept.connect(except_slot, QtCore.Qt.QueuedConnection)
        if default_exhandle:
            self.default_handler = functools.partial(default_exception_handler)  # create a copy of the function
            self.emitter.sigExcept.connect(self.default_handler, QtCore.Qt.QueuedConnection)

    def run(self):
        """
        Override virtual run method of QRunnable base class to run the function/method that was passed in the
        constructor
        """

        # print 'Started {0} in thread {1}, will emit back to {2}'.format(self._method.__name__,
        #                                                                 QtCore.QThread.currentThread(),
        #                                                                 self._callback_slot.__name__)
        try:
            if self.lock is not None:
                self.lock.lock()
            value = self._method(*self.method_args, **self.method_kwargs)
            try:
                if self._callback_slot: self.emit(self._callback_slot, value)
            except RuntimeError:
                etype, ex, tb = sys.exc_info()
                print 'exception1:', etype, ex, traceback.format_exc(), self._callback_slot, self._method
                print 'this did not run'
                msg.logMessage(('Runnable method tried to return value, but signal was already disconnected.'),
                               msg.WARNING)
                if self.lock is not None: self.lock.unlock()
                return

        except Exception:
            etype, ex, tb = sys.exc_info()
            print 'exception:',etype,ex,traceback.format_exc(),self._callback_slot,self._method
            self.emitter.sigExcept.emit(etype, ex, tb)
        else:
            self.emitter.sigFinished.emit()
        finally:
            if self.lock is not None:
                self.lock.unlock()

    def emit(self,slot,value):
        if slot is None: return
        if type(value) is not tuple: value = (value,)
        if str(type(slot)) == "<type 'PySide.QtCore.SignalInstance'>": # allows slotting into signals; this type is not in QtCore, so must compare by name str
            if slot is None: return
            value = map(nonesigmod.pyside_none_wrap, value)
            tempemitter = EmitterFactory(*[object] * len(value))()
            tempemitter.sigTemp.connect(slot, QtCore.Qt.QueuedConnection)
            tempemitter.sigTemp.emit(*value)
        else:
            guiinvoker.invoke_in_main_thread(slot, *value) # actually works better than communicating with signals





class RunnableIterator(RunnableMethod):
    """
    Runnable that will loop through an iterator and emit a signal representing the progress of a
    iterator or generator method.

    See RunnableMethod for attributes


    Attributes
    ----------
    interrupt_signal : QtCore.Signal
        Signal used to interrupt generator


    Parameters
    ----------
    iterator : iterator
        Generator/iterator object to loop through on seperate thread
    iterator_args : list
        List/tuple of arguments for the given function/method
    iterator_kwargs : dict
        Dictionary with keyword arguments for the give function/method
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    interrupt_signal : QtCore.Signal
        Signal used to interrupt generator
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    default_exhandle : bool
        Flag to use the default exception handle slot. If false it will not be called
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented
    parent : object
        Parent object reference; keeps the parent from being garbage collected before a callback. If the parent is ready
        for GC'ing, the iterator will stop, and should free the final reference.
    """

    def __init__(self, iterator, iterator_args=(), iterator_kwargs={}, callback_slot=None,
                 finished_slot=None, interrupt_signal=None, except_slot=None, default_exhandle=True,
                 priority=0, lock=None, parent=None):
        super(RunnableIterator, self).__init__(method=iterator, method_args=iterator_args,
                                               method_kwargs=iterator_kwargs, callback_slot=callback_slot,
                                               finished_slot=finished_slot, except_slot=except_slot,
                                               default_exhandle=default_exhandle, priority=priority, lock=lock)
        self._interrupt = False
        self._callback_slot = callback_slot
        self._parent = parent

        if interrupt_signal is not None:
            interrupt_signal.connect(self.interrupt)

    def interrupt(self):
        self._interrupt = True

    def run(self):
        """
        Override virtual run method of QRunnable base class
        """

        # print 'Started {0} in thread {1}, will update to {2}'.format(self._method.__name__,
        #                                                              QtCore.QThread.currentThread(),
        #                                                              self._callback_slot.__name__)
        if self.lock is not None: self.lock.lock()
        try:
            for status in self._method(*self.method_args, **self.method_kwargs):
                if self._interrupt:
                    raise StopIteration('{0} running in background thread {1} interrupted'.format(
                                            self._method.__name__, QtCore.QThread.currentThreadId()))
                try:
                    # print 'status:',status
                    self.emit(self._callback_slot,status)
                    # self.emitter.sigRetValue.emit(status)
                except RuntimeError:
                    etype, ex, tb = sys.exc_info()
                    print 'exception1:', etype, ex, traceback.format_exc(), self._callback_slot, self._method
                    print 'this did not run'
                    msg.logMessage(('Runnable iterator tried to return value, but signal was already disconnected.'),msg.WARNING)
                    if self.lock is not None: self.lock.unlock()

        except Exception:
            etype, ex, tb = sys.exc_info()
            self.emitter.sigExcept.emit(etype, ex, tb)
        else:
            self.emitter.sigFinished.emit()
        finally:
            if self.lock is not None: self.lock.unlock()


def method(callback_slot=None, finished_slot=None, except_slot=None, default_exhandle=True, lock=None):
    """
    Decorator for functions/methods to run as RunnableMethods on background QT threads
    Use it as any python decorator to decorate a function with @decorator syntax or at runtime:
    decorated_method = threads.method(callback_slot, ...)(method_to_decorate)
    then simply run it: decorated_method(*args, **kwargs)

    Parameters
    ----------
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    default_exhandle : bool
        Flag to use the default exception handle slot. If false it will not be called
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented

    Returns
    -------
    wrap_runnable_method : function
        Decorated function/method
    """

    def wrap_runnable_method(function):
        @functools.wraps(function)
        def _runnable_method(*args, **kwargs):
            runnable = RunnableMethod(function, method_args=args, method_kwargs=kwargs,
                                      callback_slot=callback_slot, finished_slot=finished_slot,
                                      except_slot=except_slot, default_exhandle=default_exception_handler, lock=lock)
            add_to_queue(runnable)
        return _runnable_method
    return wrap_runnable_method


def iterator(callback_slot=None, finished_slot=None, interrupt_signal=None, except_slot=None, lock=None, parent=None):
    """
    Decorator for iterators/generators to run as RunnableIterators on background QT threads
    Use it as any python decorator to decorate a function with @decorator syntax or at runtime:
    decorated_iterator = threads.iterator(callback_slot, ...)(iterator_to_decorate).
    then simply run it: decorated_iterator(*args, **kwargs)

    Parameters
    ----------
    callback_slot : function
        Function/method to run on a background thread
    finished_slot : QtCore.Slot
        Slot to call with the return value of the function
    interrupt_signal : QtCore.Signal
        Signal to break out of iterator loop prematurely
    except_slot : QtCore.Slot
        Function object (qt slot), slot to receive exception type, instance and traceback object
    lock : mutex/semaphore
        Simple lock if multiple access needs to be prevented

    Returns
    -------
    wrap_runnable_iterator : function
        Decorated iterator/generator
    """

    def wrap_runnable_iterator(generator):
        @functools.wraps(generator)
        def _runnable_iterator(*args, **kwargs):
            runnable = RunnableIterator(generator, iterator_args=args, iterator_kwargs=kwargs,
                                        callback_slot=callback_slot, finished_slot=finished_slot,
                                        interrupt_signal=interrupt_signal, except_slot=except_slot,
                                        lock=lock, parent=parent)
            add_to_queue(runnable)
        return _runnable_iterator
    return wrap_runnable_iterator


class Worker(QtCore.QThread):
    """
    Daemon worker that contains a Queue and QThreadPool for running jobs

    Attributes
    ----------
    queue : Queue
        Queue to put/get runnables
    pool : QThreadPool
        Thread pool used to run runnables from queue

    Parameters
    ----------
    queue : Queue
        Queue to put/get runnables
    """

    def __init__(self, queue, parent=None):
        super(Worker, self).__init__(parent)
        self.queue = queue
        self.pool = QtCore.QThreadPool.globalInstance()
        print 'maxthreads:',self.pool.maxThreadCount()


    # def __del__(self):
    #     self.queue.join()

    def stop(self):
        """
        Puts a None in the queue to signal the run loop to exit
        """

        self.queue.put(None)
        self.quit()
        print 'threads:',self.pool.activeThreadCount()

    def run(self):
        """
        Continuously get Runnables from queue and running them on available threads as long as stop flag is false
        """
        while True:
            runnable = self.queue.get()
            # print "Worker got item {} off queue".format(type(runnable))
            if runnable is None:
                break
            self.pool.start(runnable, runnable._priority)
            self.queue.task_done()


@QtCore.Slot(type, Exception, types.TracebackType)
def default_exception_handler(etype, ex, tb):
    """
    Default exception handle slot that will log the exception and traceback to the logger from pipeline.msg
    this will also print the exception

    Parameters
    ----------
    etype : Exception type
        Type of the exception being raised from background thread
    ex : Exception
        Exception object raised from the function sent to a background thread
    tb : Traceback
        Traceback object raised from the function sent to a background thread

    Returns
    -------
    None
    """

    exmessage = ''.join(traceback.format_exception(etype, ex, tb))  # concatenate the list of strings
    msg.logMessage(exmessage, level=msg.ERROR)
    traceback.print_exception(etype, ex, tb)


def add_to_queue(runnable):
    """
    Add a Runnable to the global thread.queue

    Parameters
    ----------
    runnable : QtCore.Runnable

    """
    global queue
    try:
        queue.put(runnable)
    except Exception as e:
        msg.logMessage(e, level=msg.ERROR)
        raise e


# Application globals
queue = Queue.Queue()
worker = Worker(queue)
mutex = QtCore.QMutex()
worker.start()



import pysftp


class SFTPClient(pysftp.Connection):
    """Save the host name of a pysftp Connection"""

    def __init__(self, host, **kwargs):
        super(SFTPClient, self).__init__(host, **kwargs)
        self.host = host



import tomopy
import argparse
import numpy as np
import afnumpy as afnp
import arrayfire as af
from gnufft import tvd_update,add_hessian

from XT_ForwardModel import forward_project, init_nufft_params, back_project

def gpuGridrec(tomo,angles,center,input_params):
        """
        Gridrec reconstruction using GPU based gridding
        Inputs: tomo : 3D numpy sinogram array with dimensions same as tomopy
        angles : Array of angles in radians
        center : Floating point center of rotation
        input_params : A dictionary with the keys
        'gpu_device' : Device id of the gpu (For a 4 GPU cluster ; 0-3)
        'oversamp_factor': A factor by which to pad the image/data for FFT
        'fbp_filter_param' : A number between 0-1 for setting the filter cut-off for FBP
        """

        print('Starting GPU NUFFT recon')
        #allocate space for final answer 
        af.set_device(input_params['gpu_device']) #Set the device number for gpu based code
        #Change tomopy format
        new_tomo=np.transpose(tomo,(1,2,0)) #slice, columns, angles
        im_size =  new_tomo.shape[1]
        num_slice = new_tomo.shape[0]
        num_angles=new_tomo.shape[2]
        pad_size=np.int16(im_size*input_params['oversamp_factor'])
#        nufft_scaling = (np.pi/pad_size)**2
        #Initialize structures for NUFFT
        sino={}
        geom={}
        sino['Ns'] =  pad_size#Sinogram size after padding
        sino['Ns_orig'] = im_size #size of original sinogram
        sino['center'] = center + (sino['Ns']/2 - sino['Ns_orig']/2)  #for padded sinogram
        sino['angles'] = angles
        sino['filter'] = input_params['fbp_filter_param'] #Paramter to control strength of FBP filter normalized to [0,1]

        #Initialize NUFFT parameters
        nufft_params = init_nufft_params(sino,geom)
        rec_nufft = afnp.zeros((num_slice/2,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)
        Ax = afnp.zeros((sino['Ns'],num_angles),dtype=afnp.complex64)
        pad_idx = slice(sino['Ns']/2-sino['Ns_orig']/2,sino['Ns']/2+sino['Ns_orig']/2)
        rec_nufft_final=np.zeros((num_slice,sino['Ns_orig'],sino['Ns_orig']),dtype=np.float32)
        
        #Move all data to GPU
        slice_1=slice(0,num_slice,2)
        slice_2=slice(1,num_slice,2)
        gdata=afnp.array(new_tomo[slice_1]+1j*new_tomo[slice_2],dtype=afnp.complex64)
        x_recon = afnp.zeros((sino['Ns'],sino['Ns']),dtype=afnp.complex64)
        #loop over all slices
        for i in range(0,num_slice/2):
          Ax[pad_idx,:]=gdata[i]
          #filtered back-projection 
          rec_nufft[i] = (back_project(Ax,nufft_params))[pad_idx,pad_idx]


        #Move to CPU
        #Rescale result to match tomopy
        rec_nufft=np.array(rec_nufft,dtype=np.complex64) #*nufft_scaling
        rec_nufft_final[slice_1]=np.array(rec_nufft.real,dtype=np.float32)
        rec_nufft_final[slice_2]=np.array(rec_nufft.imag,dtype=np.float32)
        return rec_nufft_final


def gpuSIRT(tomo,angles,center,input_params):
        """
        SIRT reconstruction using GPU based gridding operators
        Inputs: tomo : 3D numpy sinogram array with dimensions same as tomopy
        angles : Array of angles in radians
        center : Floating point center of rotation
        input_params : A dictionary with the keys
        'gpu_device' : Device id of the gpu (For a 4 GPU cluster ; 0-3)
        'oversamp_factor': A factor by which to pad the image/data for FFT
        'num_iter' : Number of SIRT iterations
        """
        print('Starting GPU SIRT recon')
        #allocate space for final answer 
        af.set_device(input_params['gpu_device']) #Set the device number for gpu based code
        #Change tomopy format
        new_tomo=np.transpose(tomo,(1,2,0)) #slice, columns, angles
        im_size =  new_tomo.shape[1]
        num_slice = new_tomo.shape[0]
        num_angles=new_tomo.shape[2]
        pad_size=np.int16(im_size*input_params['oversamp_factor'])
#        nufft_scaling = (np.pi/pad_size)**2
        num_iter = input_params['num_iter']
        #Initialize structures for NUFFT
        sino={}
        geom={}
        sino['Ns'] =  pad_size#Sinogram size after padding
        sino['Ns_orig'] = im_size #size of original sinogram
        sino['center'] = center + (sino['Ns']/2 - sino['Ns_orig']/2)  #for padded sinogram
        sino['angles'] = angles
        
        #Initialize NUFFT parameters
        nufft_params = init_nufft_params(sino,geom)
        temp_y = afnp.zeros((sino['Ns'],num_angles),dtype=afnp.complex64)
        temp_x = afnp.zeros((sino['Ns'],sino['Ns']),dtype=afnp.complex64)
        x_recon  = afnp.zeros((num_slice/2,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64) 
        pad_idx = slice(sino['Ns']/2-sino['Ns_orig']/2,sino['Ns']/2+sino['Ns_orig']/2)

        #allocate output array
        rec_sirt_final=np.zeros((num_slice,sino['Ns_orig'],sino['Ns_orig']),dtype=np.float32)

        #Pre-compute diagonal scaling matrices ; one the same size as the image and the other the same as data
        #initialize an image of all ones
        x_ones= afnp.ones((sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)
        temp_x[pad_idx,pad_idx]=x_ones
        temp_proj=forward_project(temp_x,nufft_params) #*(sino['Ns']*afnp.pi/2)
        R = 1/afnp.abs(temp_proj)
        R[afnp.isnan(R)]=0
        R[afnp.isinf(R)]=0
        R=afnp.array(R,dtype=afnp.complex64)
 
        #Initialize a sinogram of all ones
        y_ones=afnp.ones((sino['Ns_orig'],num_angles),dtype=afnp.complex64)
        temp_y[pad_idx]=y_ones
        temp_backproj=back_project(temp_y,nufft_params) #*nufft_scaling/2
        C = 1/(afnp.abs(temp_backproj))
        C[afnp.isnan(C)]=0
        C[afnp.isinf(C)]=0
        C=afnp.array(C,dtype=afnp.complex64)
        
        #Move all data to GPU
        slice_1=slice(0,num_slice,2)
        slice_2=slice(1,num_slice,2)
        gdata=afnp.array(new_tomo[slice_1]+1j*new_tomo[slice_2],dtype=afnp.complex64)
        
        #loop over all slices
        for i in range(num_slice/2):
          for iter_num in range(num_iter):
            #filtered back-projection
            temp_x[pad_idx,pad_idx]=x_recon[i]
            Ax = forward_project(temp_x,nufft_params)
            temp_y[pad_idx]=gdata[i]
            x_recon[i] = x_recon[i]+(C*back_project(R*(temp_y-Ax),nufft_params))[pad_idx,pad_idx] #nufft_scaling

        #Move to CPU
        #Rescale result to match tomopy
        rec_sirt=np.array(x_recon,dtype=np.complex64)
        rec_sirt_final[slice_1]=np.array(rec_sirt.real,dtype=np.float32)
        rec_sirt_final[slice_2]=np.array(rec_sirt.imag,dtype=np.float32)
        return rec_sirt_final


def gpuMBIR(tomo,angles,center,input_params):
        """
        MBIR reconstruction using GPU based gridding operators
        Inputs: tomo : 3D numpy sinogram array with dimensions same as tomopy
        angles : Array of angles in radians
        center : Floating point center of rotation
        input_params : A dictionary with the keys
        'gpu_device' : Device id of the gpu (For a 4 GPU cluster ; 0-3)
        'oversamp_factor': A factor by which to pad the image/data for FFT
        'num_iter' : Max number of MBIR iterations
        'smoothness' : Regularization constant
        'p': MRF shape param
        """
        print('Starting GPU MBIR recon')
        #allocate space for final answer 
        af.set_device(input_params['gpu_device']) #Set the device number for gpu based code
        #Change tomopy format
        new_tomo=np.transpose(tomo,(1,2,0)) #slice, columns, angles
        im_size =  new_tomo.shape[1]
        num_slice = new_tomo.shape[0]
        num_angles=new_tomo.shape[2]
        pad_size=np.int16(im_size*input_params['oversamp_factor'])
#        nufft_scaling = (np.pi/pad_size)**2
        num_iter = input_params['num_iter']
        mrf_sigma = input_params['smoothness']
        mrf_p = input_params['p']
        print('MRF params p=%f sigma=%f' %(mrf_p,mrf_sigma))
        #Initialize structures for NUFFT
        sino={}
        geom={}
        sino['Ns'] =  pad_size#Sinogram size after padding
        sino['Ns_orig'] = im_size #size of original sinogram
        sino['center'] = center + (sino['Ns']/2 - sino['Ns_orig']/2)  #for padded sinogram
        sino['angles'] = angles
        
        #Initialize NUFFT parameters
        print('Initialize NUFFT params')
        nufft_params = init_nufft_params(sino,geom)

        temp_y = afnp.zeros((sino['Ns'],num_angles),dtype=afnp.complex64)
        temp_x = afnp.zeros((sino['Ns'],sino['Ns']),dtype=afnp.complex64)
        x_recon  = afnp.zeros((num_slice/2,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)
        
        pad_idx = slice(sino['Ns']/2-sino['Ns_orig']/2,sino['Ns']/2+sino['Ns_orig']/2)

        #allocate output array
        rec_mbir_final=np.zeros((num_slice,sino['Ns_orig'],sino['Ns_orig']),dtype=np.float32)
        
        #Move all data to GPU
        print('Moving data to GPU')
        slice_1=slice(0,num_slice,2)
        slice_2=slice(1,num_slice,2)
        gdata=afnp.array(new_tomo[slice_1]+1j*new_tomo[slice_2],dtype=afnp.complex64)
        gradient = afnp.zeros((num_slice/2,sino['Ns_orig'],sino['Ns_orig']), dtype=afnp.complex64)#temp array to store the derivative of cost func
        z_recon  = afnp.zeros((num_slice/2,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)#Nesterov method variables
        t_nes = 1
        
        #Compute Lipschitz of gradient
        print('Computing Lipschitz of gradient')
        x_ones= afnp.ones((1,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)
        temp_x[pad_idx,pad_idx]=x_ones[0]
        temp_proj=forward_project(temp_x,nufft_params)
        temp_backproj=(back_project(temp_proj,nufft_params))[pad_idx,pad_idx]
        print('Adding Hessian of regularizer')
        temp_backproj2=afnp.zeros((1,sino['Ns_orig'],sino['Ns_orig']),dtype=afnp.complex64)
        temp_backproj2[0]=temp_backproj
        add_hessian(mrf_sigma,x_ones, temp_backproj2)
        L = np.max([temp_backproj2.real.max(),temp_backproj2.imag.max()])
        print('Lipschitz constant = %f' %(L))
        del x_ones,temp_proj,temp_backproj,temp_backproj2

        #loop over all slices
        for iter_num in range(num_iter):
          print('Iteration %d of %d'%(iter_num,num_iter))
        #Derivative of the data fitting term
          for i in range(num_slice/2):
            temp_x[pad_idx,pad_idx]=x_recon[i]
            Ax = forward_project(temp_x,nufft_params)
            temp_y[pad_idx]=gdata[i]
            gradient[i] =(back_project((Ax-temp_y),nufft_params))[pad_idx,pad_idx] #nufft_scaling
        #Derivative of regularization term
          tvd_update(mrf_p,mrf_sigma,x_recon, gradient) 
          #x_recon-=gradient/L
          x_recon,z_recon,t_nes=nesterovOGM2update(x_recon,z_recon,t_nes,gradient,L)
        
        #Move to CPU
        #Rescale result to match tomopy
        rec_mbir=np.array(x_recon,dtype=np.complex64)
        rec_mbir_final[slice_1]=np.array(rec_mbir.real,dtype=np.float32)
        rec_mbir_final[slice_2]=np.array(rec_mbir.imag,dtype=np.float32)
        return rec_mbir_final

def nesterovOGM1update(x,z,t,grad,L):
#L = Lipshcitz constant
    tNew = 0.5*(1+np.sqrt(1+4*(t**2)))
    zNew = x - grad/L
    xNew = zNew + ((t-1)/tNew)*(zNew-z)
    return xNew,zNew,tNew

def nesterovOGM2update(x,z,t,grad,L):
#L = Lipshcitz constant
    zNew = x - grad/L
    tNew = 0.5*(1+np.sqrt(1+4*(t**2)))
    xNew =zNew + ((t -1)/tNew)*(zNew-z) + (t/tNew)*(zNew-x)
    return xNew,zNew,tNew
                                



#Fuction to implement the tomographic forward and back-projection kernel in python

import gnufft 
import math
import numpy as np
import afnumpy as afnp 
import afnumpy.fft as af_fft
import scipy.special as sc_spl #For bessel functions
import tomopy
import matplotlib.pyplot as plt
from XT_Common import padmat

def forward_project(x,params):
    #inputs : x - afnumpy array containing the complex valued image
    #       : params - a list containing all parameters for the NUFFT 

    qxyXrxy = (params['fft2Dshift']*af_fft.fft2(x*params['deapod_filt']*params['fft2Dshift'])) #real space (rxy) to Fourier space (qxy)
    
    qtXqxy = gnufft.polarsample(params['gxy'],qxyXrxy,params['gkblut'],params['scale'],params['k_r'])/(params['Ns']**2) #Fourier space to polar coordinates interpolation (qxy to qt)

    rtXqt = params['fftshift1D']((af_fft.ifft(afnp.array(params['fftshift1D_center'](qtXqxy).T))).T)*params['sino_mask'] #Polar cordinates to real space qt to rt
    #TODO : Remove afnp array allocation

    return rtXqt 

def back_project(y,params):
    #inputs : y - afnumpy array containing the complex valued array with size of the sinogram 
    #       : params - a list containing all parameters for the NUFFT 

    qtXrt = params['giDq'].reshape((params['Ns'],1))*(params['fftshift1Dinv_center'](af_fft.fft((params['fftshift1D'](y)).T).T)) #Detector space rt to Fourier space qt

    #Polar to cartesian 
#    qxyXqt = gnufft.polarsample_transpose(params['gxy'],qtXrt,params['grid'],params['gkblut'],params['scale'],params['k_r']) *(afnp.pi/(2*params['Ntheta']*params['Ns']**2))
    qxyXqt = gnufft.polarsample_transpose(params['gxy'],qtXrt,params['grid'],params['gkblut'],params['scale'],params['k_r']) *(afnp.pi/(2*params['Ns']))

    rxyXqxy =params['fft2Dshift']*(af_fft.ifft2(qxyXqt*params['fft2Dshift']))*params['deapod_filt'] #Fourier to real space : qxy to rxy

    return rxyXqxy


def init_nufft_params(sino,geom):
   # Function to initialize parameters associated with the forward model 
    #inputs : sino - A list contating parameters associated with the sinogram 
    #              Ns : Number of entries in the padded sinogram along the "detector" rows 
    #              Ns_orig :  Number of entries  detector elements per slice
    #              center : Center of rotation in pixels computed from the left end of the detector
    #              angles : An array containg the angles at which the data was acquired in radians
    #       : geom - TBD
    #
     
    KBLUT_LENGTH = 256
    k_r=3 #kernel size 2*kr+1 TODO : Breaks when k_r is large. Why ? 
    beta =4*math.pi  
    Ns = sino['Ns']
    Ns_orig = sino['Ns_orig']
    ang = sino['angles']

    q_grid = np.arange(1,sino['Ns']+1) - np.floor((sino['Ns']+1)/2) - 1
    sino['tt'],sino['qq']=np.meshgrid(ang*180/math.pi,q_grid)

    # Preload the Bessel kernel (real components!)
    kblut,KB,KB1D,KB2D=KBlut(k_r,beta,KBLUT_LENGTH) 


    #Normalization (density compensation factor)
#    Dq=KBdensity1(sino['qq'],sino['tt'],KB1,k_r,Ns)';

    # polar to cartesian, centered
    [xi,yi]=pol2cart(sino['qq'],sino['tt']*math.pi/180)
    xi = xi+np.floor((Ns+1)/2)
    yi = yi+np.floor((Ns+1)/2)
   
    params={}
    params['k_r'] = k_r
    params['deapod_filt']=afnp.array(deapodization(Ns,KB1D),dtype=afnp.float32)
    params['sino_mask'] = afnp.array(padmat(np.ones((Ns_orig,sino['qq'].shape[1])),np.array((Ns,sino['qq'].shape[1])),0),dtype=afnp.float32)
    params['grid'] = [Ns,Ns] 
    params['scale']= ((KBLUT_LENGTH-1)/k_r)
    params['center'] = afnp.array(sino['center'])
    params['Ns'] = Ns
    params['Ntheta'] = np.size(ang)

    # push parameters to gpu and initalize a few in-line functions 
    params['gxi'] = afnp.array(np.single(xi))
    params['gyi'] = afnp.array(np.single(yi))
    params['gxy'] = params['gxi']+1j*params['gyi']
    params['gkblut'] = afnp.array(np.single(kblut))
    params['det_grid'] = np.array(np.reshape(np.arange(0,sino['Ns']),(sino['Ns'],1)))

    #####Generate Ram-Lak/ShepLogan like filter kernel#########
    
    temp_mask=np.ones(Ns)
    kernel=np.ones(Ns)
    if 'filter' in sino:
      temp_r = np.linspace(-1,1,Ns)
      kernel = (Ns)*np.fabs(temp_r)*np.sinc(temp_r/2)
      temp_pos = (1-sino['filter'])/2
      temp_mask[0:np.int16(temp_pos*Ns)]=0
      temp_mask[np.int16((1-temp_pos)*Ns):]=0
    params['giDq']=afnp.array(kernel*temp_mask,dtype=afnp.complex64)
    
    temp = afnp.array((-1)**params['det_grid'],dtype=afnp.float32)
    temp2 = np.array((-1)**params['det_grid'],dtype=afnp.float32)
    temp2 = afnp.array(temp2.reshape(1,sino['Ns']))
    temp3 = afnp.array(afnp.exp(-1j*2*params['center']*(afnp.pi/params['Ns'])*params['det_grid']).astype(afnp.complex64))
    temp4 = afnp.array(afnp.exp(1j*2*params['center']*afnp.pi/params['Ns']*params['det_grid']).astype(afnp.complex64))
    params['fft2Dshift'] = afnp.array(temp*temp2,dtype=afnp.complex64)
    params['fftshift1D'] = lambda x : temp*x
    params['fftshift1D_center'] = lambda x : temp3*x
    params['fftshift1Dinv_center'] = lambda x : temp4*x
    
    return params

def deapodization(Ns,KB1D):
    xx=np.arange(1,Ns+1)-Ns/2-1
    dpz=np.fft.fftshift(np.fft.ifft2(np.fft.fftshift(np.reshape(KB1D(xx),(np.size(xx),1))*KB1D(xx))))
    dpz=dpz.real #astype(float)
    dpz=1/dpz         
    return dpz
    
def KBlut(k_r,beta,nlut):
    kk=np.linspace(0,k_r,nlut)
    kblut = KB2( kk, 2*k_r, beta)
    scale = (nlut-1)/k_r
    kbcrop = lambda x: (np.abs(x)<=k_r)
    KBI = lambda x: np.int16(np.abs(x)*scale-np.floor(np.abs(x)*scale))
    KB1D = lambda x: (np.reshape(kblut[np.int16(np.floor(np.abs(x)*scale)*kbcrop(x))],x.shape)*KBI(x)+np.reshape(kblut[np.int16(np.ceil(np.abs(x)*scale)*kbcrop(x))],x.shape)*(1-KBI(x)))*kbcrop(x)
    KB=lambda x,y: KB1D(x)*KB1D(y)
    KB2D=lambda x,y: KB1D(x)*KB1D(y)
    return kblut, KB, KB1D,KB2D

def KB2(x, k_r, beta):
    w = sc_spl.iv(0, beta*np.sqrt(1-(2*x/k_r)**2)) 
    #w=w/np.abs(sc_spl.iv(0, beta))
    w=(w*(x<=k_r))
    return w

def cart2pol(x, y):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)
    return(rho, phi)

def pol2cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return(x, y)

#def densityCompensation(qq,tt,KB,nj,Ns):
#    nb=100; #TODO : Why 100 ? Venkat
#    [nt,nq]=size(qq);   
    #crop repeated angles
#    ii=(tt(:,1))-min(tt(:,1))<180;
#    qq1=qq(ii,:);
#    tt1=tt(ii,:);
#    qq1=-fliplr(qq1);
#    [xi,yi]=pol2cart(tt1*pi/180,qq1);



from PySide import QtGui, QtCore
from fabio import edfimage, tifimage
import numpy as np
import os

def savedatadialog(guesspath="", caption="Save data to EDF", headers=None):
    if headers is None:
        headers = dict()

    dialog = QtGui.QFileDialog(parent=None, caption=caption, directory=os.path.dirname(guesspath),
                               filter=u"EDF (*.edf);;PNG (*.png);;TIFF (*.tif)")
    dialog.selectFile(os.path.basename(guesspath))
    filename, ok = dialog.getSaveFileName()



    return filename, ok

def infodialog(msg,shortmsg):
    msgBox = QtGui.QMessageBox()
    msgBox.setText(msg)
    msgBox.setInformativeText(shortmsg)
    msgBox.setStandardButtons(QtGui.QMessageBox.Close)
    msgBox.setDefaultButton(QtGui.QMessageBox.Close)

    response = msgBox.exec_()

def checkoverwrite(path):
    msgBox = QtGui.QMessageBox()
    msgBox.setText('Are you sure you want to overwrite {}?'.format(path))
    # msgBox.setInformativeText(shortmsg)
    msgBox.setStandardButtons(QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel)
    msgBox.setDefaultButton(QtGui.QMessageBox.Cancel)

    response = msgBox.exec_()
    return response==QtGui.QMessageBox.Ok


from PySide import QtCore, QtGui
from PySide.QtUiTools import QUiLoader

particlemenu = None
blankForm = None
leftwidget = None
centerwidget = None
rightwidget = None


def load():
    global leftwidget, centerwidget, rightwidget, blankForm
    # Load the gui from file
    guiloader = QUiLoader()
    f = QtCore.QFile("xicam/gui/hipgisaxsleftwidget.ui")
    f.open(QtCore.QFile.ReadOnly)
    leftwidget = guiloader.load(f)
    f.close()

    rightwidget = QtGui.QStackedWidget()
    centerwidget = QtGui.QSplitter(QtCore.Qt.Vertical)

    blankForm = QtGui.QLabel('Select a feature on the left panel to edit...')
    blankForm.setSizePolicy(QtGui.QSizePolicy.Ignored, QtGui.QSizePolicy.Ignored)
    blankForm.setAlignment(QtCore.Qt.AlignCenter)
    showForm(blankForm)

    #leftmodes = [(leftwidget, QtGui.QFileIconProvider().icon(QtGui.QFileIconProvider.File))]

    return centerwidget, rightwidget, leftwidget


def showForm(form):
    rightwidget.addWidget(form)
    rightwidget.setCurrentWidget(form)


from PySide import QtGui, QtCore


__author__ = "Luis Barroso-Luque, Ronald J Pandolfi"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


class FeatureWidget(QtGui.QWidget):
    """
    Widget that stands for a feature or function with a a preview icon and a delete icon. These are added to a layout
    to represent a set of features (as in higgisaxs) or functions (as in tomography). FeatureWidgets are normally
    managed by an instance of the FeatureManager Class


    Attributes
    ----------
    name : str
        Name to be shown in the FeatureWidget's GUI
    form
        Widget that portrays information about the widget (ie a textlabel, spinbox, ParameterTree, etc)
    subfeatures : list of FeatureWidgets/QWidgets
        Subfeatures of the widget
    previewButton : QtGui.QPushButton
        Button to call the preview action associated with this widget
    closeButton : QtGui.QPushButton
        Button that emits sigDelete

    Signals
    -------
    sigClicked(FeatureWidget)
        Signal emitted when the widget is clicked, self is emitted so that a FeatureManager has access to the sender
        Qt has a sender method which could also be used instead...
    sigDelete(FeatureWidget)
        Signal emitted when the widget's closeButton is clicked, self is emitted
    sigSubFeature(FeatureWidget)
        Signal emitted when a subfeature is added to the current widget. Emits the subfeature.


    Parameters
    ----------
    name : str
        Name to be given to widget
    checkable : bool, optional
        Boolean specifying if the previewButton is checkable (for toggling)
    closeable : bool, optional
        Boolean specifying whether the widget can be closed/deleted
    subfeatures : list of FeatureWidgets/QWidgets
        Initialization list of subfeatures. New subfeatures can be added with the addSubFeatureMethod
    parent
        Parent widget, normally the manager
    """

    sigClicked = QtCore.Signal(QtGui.QWidget)
    sigDelete = QtCore.Signal(QtGui.QWidget)
    sigSubFeature = QtCore.Signal(QtGui.QWidget)

    def __init__(self, name='', checkable=True, closeable=True, subfeatures=None, parent=None):
        super(FeatureWidget, self).__init__(parent=parent)

        self.name = name
        self.form = QtGui.QLabel(self.name) # default form
        self.subfeatures = []

        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)
        self.verticalLayout = QtGui.QVBoxLayout(self)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.frame = QtGui.QFrame(self)
        self.frame.setFrameShape(QtGui.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtGui.QFrame.Raised)
        self.horizontalLayout_2 = QtGui.QHBoxLayout(self.frame)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.previewButton = QtGui.QPushButton(parent=self)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.previewButton.sizePolicy().hasHeightForWidth())
        self.previewButton.setSizePolicy(sizePolicy)
        self.previewButton.setStyleSheet("margin:0 0 0 0;")
        self.previewButton.setText("")
        icon = QtGui.QIcon()

        if checkable:
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_48.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_47.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.previewButton.setCheckable(True)
        else:
            icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_47.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
            self.previewButton.setCheckable(False)
            self.previewButton.setChecked(True)

        self.previewButton.setIcon(icon)
        self.previewButton.setFlat(True)
        self.previewButton.setChecked(True)
        self.previewButton.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.horizontalLayout_2.addWidget(self.previewButton)
        self.line = QtGui.QFrame(self.frame)
        self.line.setFrameShape(QtGui.QFrame.VLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.horizontalLayout_2.addWidget(self.line)
        self.txtName = ROlineEdit(self.frame)
        self.horizontalLayout_2.addWidget(self.txtName)
        self.txtName.setText(name)
        self.line_3 = QtGui.QFrame(self.frame)
        self.line_3.setFrameShape(QtGui.QFrame.VLine)
        self.line_3.setFrameShadow(QtGui.QFrame.Sunken)
        self.horizontalLayout_2.addWidget(self.line_3)
        self.closeButton = QtGui.QPushButton(self.frame)
        self.closeButton.setStyleSheet("margin:0 0 0 0;")
        self.closeButton.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("xicam/gui/icons_46.gif"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.closeButton.setIcon(icon1)
        self.closeButton.setFlat(True)
        self.closeButton.clicked.connect(self.delete)
        self.horizontalLayout_2.addWidget(self.closeButton)
        self.verticalLayout.addWidget(self.frame)
        self.txtName.sigClicked.connect(self.mouseClicked)
        self.frame.setFrameShape(QtGui.QFrame.Box)
        self.frame.setCursor(QtCore.Qt.ArrowCursor)

        self.subframe = QtGui.QFrame(self)
        self.subframe.setFrameShape(QtGui.QFrame.StyledPanel)
        self.subframe.setFrameShadow(QtGui.QFrame.Raised)
        self.subframe_layout = QtGui.QVBoxLayout(self.subframe)# QtGui.QGridLayout(self.subframe)
        self.subframe_layout.setContentsMargins(0, 0, 0, 0)
        self.subframe_layout.setSpacing(0)
        self.verticalLayout.addWidget(self.subframe)
        self.subframe.hide()

        if not closeable:
            self.closeButton.hide()

        if subfeatures is not None:
            for subfeature in subfeatures:
                self.addSubFeature(subfeature)

        self.collapse()

    def addSubFeature(self, subfeature):
        """
        Adds a subfeature to the widget

        Parameters
        ----------
        subfeature : FeatureWidget/QWidget
            Widget to add as a subfeature
        """

        h = QtGui.QHBoxLayout()
        indent = QtGui.QLabel('  -   ')
        h.addWidget(indent)
        subfeature.destroyed.connect(indent.deleteLater)
        subfeature.destroyed.connect(h.deleteLater)
        if isinstance(subfeature, QtGui.QLayout):
            h.addLayout(subfeature)
        elif isinstance(subfeature, QtGui.QWidget):
            h.addWidget(subfeature)
        self.subframe_layout.addLayout(h)
        try:
            subfeature.sigDelete.connect(self.removeSubFeature)
        except AttributeError:
            pass

        self.sigSubFeature.emit(subfeature)
        self.subfeatures.append(subfeature)

    def removeSubFeature(self, subfeature):
        """
        Removes a subfeature

        Parameters
        ----------
        subfeature : FeatureWidget/QWidget
            Feature to remove
        """

        self.subfeatures.remove(subfeature)
        subfeature.deleteLater()
        del subfeature

    def delete(self):
        """
        Emits delete signal with self. Connected to deleteButton's clicked
        """

        self.sigDelete.emit(self)

    def collapse(self):
        """
        Collapses all expanded subfeatures
        """

        if self.subframe is not None:
            self.subframe.hide()

    def expand(self):
        """
        Expands subfeatures
        """

        if self.subframe is not None:
            self.subframe.show()

    def mouseClicked(self):
        """
        Slot to handle when a feature is clicked
        """

        self.sigClicked.emit(self)
        self.setFocus()
        self.previewButton.setFocus()


class ROlineEdit(QtGui.QLineEdit):
    """
    Subclass of QlineEdit used for labels in FeatureWidgets
    """

    sigClicked = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super(ROlineEdit, self).__init__(*args, **kwargs)
        self.setReadOnly(True)
        self.setFrame(False)

    def focusOutEvent(self, *args, **kwargs):
        super(ROlineEdit, self).focusOutEvent(*args, **kwargs)
        self.setCursor(QtCore.Qt.ArrowCursor)

    def mousePressEvent(self, *args, **kwargs):
        super(ROlineEdit, self).mousePressEvent(*args, **kwargs)
        self.sigClicked.emit()

    def mouseDoubleClickEvent(self, *args, **kwargs):
        super(ROlineEdit, self).mouseDoubleClickEvent(*args, **kwargs)
        self.setFrame(True)
        self.setFocus()
        self.selectAll()


class FeatureManager(QtCore.QObject):
    """
    Feature Manager class to manage a list of FeatureWidgets and show the list in an appropriate layout and their
    corresponding forms in another layout. list layout must have an addWidget, removeWidget methods. Form layout
    must in addition have a setCurrentWidget method

    Attributes
    ----------
    features : list of FeatureWidgets
        List of the FeatureWidgets managed
    selectedFeature : FeatureWidget
        The currently selected feature

    Parameters
    ----------
    list_layout : QtGui.QLayout
        Layout to display the list of FeatureWidgets
    form_layout : QtGui.QLayout
        Layout to display the FeaturenWidgets form (pyqtgraph.Parameter)
    feature_widgets : list of FeatureWidgets, optional
        List with feature widgets for initialization
    blank_form : QtGui.QWidget, optional
        Widget to display in form_layout when not FunctionWidget is selected
    """

    def __init__(self, list_layout, form_layout, feature_widgets=None, blank_form=None):
        self._llayout = list_layout
        self._flayout = form_layout
        self.features = []
        self.selectedFeature = None

        if feature_widgets is not None:
            for feature in feature_widgets:
                self.addFeature(feature)

        if blank_form is not None:
            if isinstance(blank_form, str):
                self.blank_form = QtGui.QLabel(blank_form)
            else:
                self.blank_form = blank_form
        else:
            self.blank_form = QtGui.QLabel('Select a feature to view its form')
        self.blank_form.setAlignment(QtCore.Qt.AlignCenter)
        self.blank_form.setSizePolicy(QtGui.QSizePolicy.Ignored, QtGui.QSizePolicy.Ignored)

        self._flayout.addWidget(self.blank_form)
        self.showForm(self.blank_form)
        super(FeatureManager, self).__init__()

    @property
    def count(self):
        """
        Number of features managed
        """
        return len(self.features)

    @property
    def nextFeature(self):
        """
        The feature after the selectedFeature
        """
        try:
            index = self.features.index(self.selectedFeature)
        except (ValueError, IndexError):
            return None
        if index > 0:
            return self.features[index - 1]
        else:
            return self.selectedFeature

    @property
    def previousFeature(self):
        """
        The feature before the selectedFeature
        """
        if self.selectedFeature is None:
            return None
        index = self.features.index(self.selectedFeature)
        if index < self.count - 1:
            return self.features[index + 1]
        else:
            return self.selectedFeature

    @QtCore.Slot(QtGui.QWidget)
    def featureClicked(self, feature):
        """
        Slot used to receive features sigClicked
        """
        if feature in self.features:
            self.collapseAllFeatures()
        self.showForm(feature.form)
        feature.expand()
        self.selectedFeature = feature

    @QtCore.Slot(QtGui.QWidget)
    def removeFeature(self, feature):
        """
        Slot used to receive features sigDelete
        """
        self.features.remove(feature)
        feature.deleteLater()
        del feature
        self.showForm(self.blank_form)

    @QtCore.Slot(QtGui.QWidget)
    def subFeatureAdded(self, subfeature):
        """
        Slot used to receive features sigSubFeature
        """
        try:
            subfeature.sigClicked.connect(self.featureClicked)
            subfeature.sigSubFeature.connect(self.subFeatureAdded)
            self._flayout.addWidget(subfeature.form)
        except AttributeError:
            pass

    def addFeature(self, feature):
        """
        Adds a subfeature to the given feature
        """
        self.features.append(feature)
        self._llayout.addWidget(feature)
        self._flayout.addWidget(feature.form)
        feature.sigClicked.connect(self.featureClicked)
        feature.sigDelete.connect(self.removeFeature)
        feature.sigSubFeature.connect(self.subFeatureAdded)
        if feature.subfeatures is not None:
            for subfeature in feature.subfeatures:
                self.subFeatureAdded(subfeature)

    def collapseAllFeatures(self):
        """
        Collapses all features with subfeatures
        """
        for feature in self.features:
            feature.collapse()
            if feature.subfeatures is not None:
                for subfeature in feature.subfeatures:
                    subfeature.collapse()

    def showForm(self, form):
        """
        Shows the current features form
        """
        self._flayout.setCurrentWidget(form)

    def removeAllFeatures(self):
        """
        Deletes all features
        """
        for feature in self.features:
            feature.deleteLater()
            del feature
        self.features = []
        self.showForm(self.blank_form)

    def clearLayouts(self):
        """
        Removes all features and forms from layouts
        """
        for feature in self.features:
            self._flayout.removeWidget(feature)
            self._llayout.removeWidget(feature)

    def update(self):
        """
        Updates the layouts to show the current list of features
        """
        self.clearLayouts()
        for feature in self.features:
            self._llayout.addWidget(feature)
            self._flayout.addWidget(feature.form)
        self.showForm(self.selectedFeature.form)

    def swapFeatures(self, f1, f2):
        """
        Swaps the location of two features
        """
        idx_1, idx_2 = self.features.index(f1), self.features.index(f2)
        self.features[idx_1], self.features[idx_2] = self.features[idx_2], self.features[idx_1]
        self.update()
